// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_basic_move.h"

#include "strategy.h"

#include "bhv_basic_tackle.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>
#include "rcsc/action/bhv_go_to_point_look_ball.h"
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include "bhv_normal_dribble.h"
#include <rcsc/common/player_param.h>
#include <rcsc/action/intention_dribble2008.h>
#include "neck_offensive_intercept_neck.h"
#include "sample_player.h"
#include "bhv_pass_kick_find_receiver.h"
#include "rcsc/action/bhv_shoot2008.h"
#include <rcsc/action/body_clear_ball2009.h>
#include "dribble.h"
#include "domain_graph.h"
#include "pass.h"

#include <ctime>

using namespace rcsc;

/*-------------------------------------------------------------------*/
/*!

 */
bool
Bhv_BasicMove::execute( PlayerAgent * agent )
{
    dlog.addText( Logger::TEAM,
                  __FILE__": Bhv_BasicMove" );


    //-----------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.8, 80.0 ).execute( agent ) )
    {
        return true;
    }

    const WorldModel & wm = agent->world();
    /*--------------------------------------------------------*/
    // chase ball
    const int self_min = wm.interceptTable()->selfReachCycle();
    const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();


    if (! wm.existKickableTeammate()
            && ( self_min <= 3
                 || ( self_min <= mate_min
                      && self_min < opp_min + 3 )
               )
       )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": intercept" );
        Body_Intercept().execute( agent );
        agent->setNeckAction( new Neck_OffensiveInterceptNeck() );

        return true;
    }


//    if(Bhv_BasicKeepOpen().execute(agent))
//    	return true;



    const Vector2D target_point = Strategy::i().getPosition( wm.self().unum() );
    const double dash_power = Strategy::get_normal_dash_power( wm );

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    dlog.addText( Logger::TEAM,
                  __FILE__": Bhv_BasicMove target=(%.1f %.1f) dist_thr=%.2f",
                  target_point.x, target_point.y,
                  dist_thr );

    agent->debugClient().addMessage( "BasicMove%.0f", dash_power );
    agent->debugClient().setTarget( target_point );
    agent->debugClient().addCircle( target_point, dist_thr );

    if ( ! Body_GoToPoint( target_point, dist_thr, dash_power
                         ).execute( agent ) )
    {
        Body_TurnToBall().execute( agent );
    }

    if ( wm.existKickableOpponent()
            && wm.ball().distFromSelf() < 18.0 )
    {
        agent->setNeckAction( new Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new Neck_TurnToBallOrScan() );
    }

    return true;
}




bool
Bhv_BasicKeepOpen::execute( PlayerAgent * agent)
{
	dlog.addText( Logger::TEAM,
			__FILE__": Bhv_BasicKeepOpen+domain graph" );

	const WorldModel & wm = agent->world();
	//const int self_min = wm.interceptTable()->selfReachCycle();
	//const int mate_min = wm.interceptTable()->teammateReachCycle();
	//const int opp_min = wm.interceptTable()->opponentReachCycle();

	//generate a random set of reachable points
	srand(std::time(0));
	double dist=10;
	double a=0;
	double best_score=-100;
	int best_direction=0;
	Vector2D best_location(wm.self().pos().x,wm.self().pos().y);
	//std::cout<<"current location: "<<wm.self().pos().x<<" , "<<wm.self().pos().y<<std::endl;
	for(int i=0;i<20;i++)
	{
		a=rand()%360;
		double x_disp=dist*cos(a/180*3.14);
		double y_disp=dist*sin(a/180*3.14);

		Vector2D v_disp(x_disp,y_disp);	//disposition vector

		v_disp=v_disp+wm.self().pos();	//new location


		double score=evaluatePosition(wm,v_disp);
		if(score>best_score)
		{
			best_score=score;
			best_direction=a;
			best_location=v_disp;
		}
	}

	//compare best_location with the original direction of the player
	Vector2D same_old_pos(wm.self().pos().x,wm.self().pos().y);
	same_old_pos=same_old_pos+Vector2D(dist*cos(wm.self().pos().dir().radian()),dist*sin(wm.self().pos().dir().radian()));
	if(evaluatePosition(wm,same_old_pos)>0.7*best_score)
	{
		best_location=same_old_pos;
		//std::cout<<"old location is OK!"<<std::endl;
	}


	//std::cout<<wm.time().cycle()<<": new location: "<<best_location.x<<" , "<<best_location.y<<std::endl;
	const double dash_power = Strategy::get_normal_dash_power( wm );
	double dist_thr = wm.ball().distFromSelf() * 0.1;
	if ( dist_thr < 1.0 ) dist_thr = 1.0;

	if ( ! Body_GoToPoint( best_location, dist_thr, dash_power).execute( agent ) )
	{
		Body_TurnToBall().execute( agent );
	}

	if ( wm.existKickableOpponent() && wm.ball().distFromSelf() < 18.0 )
	{
		agent->setNeckAction( new Neck_TurnToBall() );
	}
	else
	{
		agent->setNeckAction( new Neck_TurnToBallOrScan() );
	}

	return true;
}


double Bhv_BasicKeepOpen::evaluatePosition(const WorldModel & wm, Vector2D v_disp)
{
    Vector2D  v_to_opp=wm.getOpponentNearestToSelf(100,true)->pos()-v_disp;
    //std::cout<<"v_to_opp: r: "<<v_to_opp.length()<<" a: "<<v_to_opp.dir().degree()<<std::endl;
    Vector2D v_to_mate=wm.getTeammateNearestToBall(100,true)->pos()-v_disp;
    //std::cout<<wm.time().cycle()<<": v_to_mate("<<(wm.getTeammateNearestToBall(100,true))->unum()<<"): r: "<<v_to_mate.length()<<" a: "<<v_to_mate.dir().degree()<<std::endl;
    double dist_to_opp=v_to_opp.length();
    double dist_to_mate=v_to_mate.length();

    AngleDeg angle_deg=v_to_opp.dir()-v_to_mate.dir();
    double angle_rad=abs(angle_deg.radian());

    //evaluate location

    //open angle -- to maximise
    double cr1=asin(dist_to_opp*sin(angle_rad)/
                    sqrt(dist_to_opp*dist_to_opp + dist_to_mate*dist_to_mate -
                         2*dist_to_opp*dist_to_mate*cos(angle_rad))
                   );


    //dist to home  -- to minimise
    double cr2=v_disp.dist(Strategy::i().getPosition( wm.self().unum() ));

    double score=0.80*cr1-0.20*cr2;

    return score;
}


/*-------------------------------------------------------------------*/
/*!
 *
 */
bool
Bhv_BasicMove::execute( PlayerAgent * agent, DomainGraph::Ptr dgraph )
{
	dlog.addText( Logger::TEAM,
			__FILE__": Bhv_BasicMove+domain graph" );

	const WorldModel & wm = agent->world();
	const int self_min = wm.interceptTable()->selfReachCycle();
	const int mate_min = wm.interceptTable()->teammateReachCycle();
	const int opp_min = wm.interceptTable()->opponentReachCycle();

	int mate_unum=-1;
	if(wm.interceptTable()->fastestTeammate())
		mate_unum=wm.interceptTable()->fastestTeammate()->unum();

	double ang_diff=0;

	if ( Bhv_BasicTackle( 0.8, 80.0 ).execute( agent ) )
	{
		return true;
	}

	//    const WorldModel & wm = agent->world();
	//    /*--------------------------------------------------------*/
	//    // chase ball
	//    const int self_min = wm.interceptTable()->selfReachCycle();
	//    const int mate_min = wm.interceptTable()->teammateReachCycle();
	//    const int opp_min = wm.interceptTable()->opponentReachCycle();


	if (! wm.existKickableTeammate()
			&& ( self_min <= 5
					|| ( self_min <= mate_min
							&& self_min < opp_min + 5 )
			)
	)
	{
		dlog.addText( Logger::TEAM,
				__FILE__": intercept" );
		Body_Intercept().execute( agent );
		agent->setNeckAction( new Neck_OffensiveInterceptNeck() );

		return true;
	}


	//        if(Bhv_BasicKeepOpen().execute(agent))
	//        	return true;
	//
	//
	//
	//        const Vector2D target_point = Strategy::i().getPosition( wm.self().unum() );

	//    if (dgraph && dgraph->action())
	//    {
	//    	if(dgraph->action()->category()==CooperativeAction::GetBall)
	//    	{
	//    		Vector2D ball_me=wm.self().pos()-wm.ball().pos();
	//    		Vector2D ball_mate=wm.getTeammateNearestToBall(100,true)->pos()-wm.ball().pos();
	//
	//    		//compute the angle between the ball moving direction and the position of self
	//    		ang_diff=fabs(ball_me.th().degree()-wm.ball().vel().th().degree());
	//    		if(ang_diff>180) ang_diff=360-ang_diff;
	//    		dlog.addText( Logger::TEAM,__FILE__": getBall -- ball-me %f ball_vel %f",ball_me.th().degree(),wm.ball().vel().th().degree());
	//
	//
	//    		if ( (self_min <= mate_min ) ||  	//it's not just kicked by me //&& ang_diff<90
	//    				(self_min > mate_min  && (ang_diff<45 || self_min-mate_min<4)) ||	//ball is not moving away from self or difference is trivial
	//    				(mate_min>100))	//no one else around
	//    		{
	//    			//if closest to ball
	//    			if (self_min <= 3  )//||&& self_min < opp_min + 5
	//    			{
	//    				dlog.addText( Logger::TEAM,__FILE__": getBall -- close to ball" );
	//
	//    				// 	CooperativeAction::ConstPtr acptr=dgraph->action();
	//    				// 	if (acptr && acptr->category()==CooperativeAction::GetBall)
	//    				// 	{
	//    				dlog.addText( Logger::TEAM,__FILE__": intercept" );
	//    				//std::cout<<"getting ball..."<<std::endl;
	//    				//Body_Intercept().execute( agent );
	//    				//agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
	//    				//Bhv_BasicTackle( 0, 80.0 ).execute( agent );
	//
	//
	//    				Vector2D ball_vel=wm.ball().vel();
	//    				ball_vel.scale(3);
	//    				//					if(Body_TackleToPoint(wm.ball().pos()+ball_vel).execute(agent)) //agent->world().ball().pos());//, const double &min_prob=0.5, const double &min_speed=0.0)
	//    				//						return true;
	//    				if(Bhv_GoToPointLookBall( wm.ball().pos()+ball_vel,10,100).execute(agent))
	//    					return true;
	//    			}
	//    			//if not closest
	//    			else
	//    			{
	//    				dlog.addText( Logger::TEAM,__FILE__": getBall -- not close to ball" );
	////    				if (self_min <= 3  )//||&& self_min < opp_min + 5
	////    					if ( Body_GoToPoint( dgraph->action()->targetPoint(),1.0,
	////    							ServerParam::i().maxDashPower() ).execute( agent ) )
	////    					{
	////    						dgraph->setAction(boost::shared_ptr<const CooperativeAction>()); //reset the action (not to learn)
	////    						agent->setNeckAction( new Neck_ScanField() );
	////
	////    						double bx=agent->world().ball().pos().x;
	////    						double by=agent->world().ball().pos().y;
	////    						double gx=ServerParam::i().ourTeamGoalPos().x;
	////    						double gy=ServerParam::i().ourTeamGoalPos().y;
	////    						//Vector2D m;
	////    						double mx=gx-bx;
	////    						double my=gy-by;
	////    						mx/=2;//sqrt(mx*mx+my*my)/2;
	////    						my/=2;//sqrt(mx*mx+my*my)/2;
	////
	////    						mx+=bx;
	////    						my+=by;
	////
	////    						//        				if(agent->effector().getPointtoPos().isValid())
	////    						//        					agent->doPointtoOff();
	////    						//        				else
	////
	////
	////    						Vector2D m(mx,my);
	////    						agent->setIntention( new IntentionGetBall(m,	// the point 2-far from the ball in our goal direction
	////    								ServerParam::i().maxDashPower(),
	////    								10,
	////    								0.9,
	////    								wm.time() ) );
	////    						return true;	//finish if an action is already selected
	////    					}
	//
	//    			}
	//    		}
	//    		else //not closest to ball
	//    		{}
	//    	}
	//    }
	//		//				if(!(self_min <= mate_min  && ang_diff<90))
	//		//dlog.addText( Logger::TEAM,__FILE__": getBall but not closest -- self_min %d mate%d_min %d ang_diff %f",self_min,mate_unum,mate_min,ang_diff);
	//		//				else if(!(self_min > mate_min  && ang_diff<45) )
	//		//					dlog.addText( Logger::TEAM,__FILE__": getBall but not closest" );
	//		//				else if (!(mate_min>100))
	//		//					dlog.addText( Logger::TEAM,__FILE__": getBall but not closest" );
	////	}
	////        if ( self_min <= mate_min  )
	////        {
	////            if (self_min <= 3  )//||&& self_min < opp_min + 5
	////            {
	////// 	CooperativeAction::ConstPtr acptr=dgraph->action();
	////// 	if (acptr && acptr->category()==CooperativeAction::GetBall)
	////// 	{
	////                dlog.addText( Logger::TEAM,
	////                              __FILE__": intercept" );
	////                std::cout<<"getting ball..."<<std::endl;
	////                //Body_Intercept().execute( agent );
	////                //agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
	////                //Bhv_BasicTackle( 0, 80.0 ).execute( agent );
	////                Body_TackleToPoint(agent->world().ball().pos());//, const double &min_prob=0.5, const double &min_speed=0.0)
	////                return true;
	////            }
	////            else
	////            {
	////                if ( Body_GoToPoint( dgraph->action()->targetPoint(),
	////                                     1.0,
	////                                     ServerParam::i().maxDashPower() ).execute( agent ) )
	////                {
	////		    dgraph->setAction(boost::shared_ptr<const CooperativeAction>()); //reset the action (not to learn)
	////                    agent->setNeckAction( new Neck_ScanField() );
	////
	////		    double bx=agent->world().ball().pos().x;
	////		    double by=agent->world().ball().pos().y;
	////		    double gx=ServerParam::i().ourTeamGoalPos().x;
	////		    double gy=ServerParam::i().ourTeamGoalPos().y;
	////		    //Vector2D m;
	////		    double mx=gx-bx;
	////		    double my=gy-by;
	////		    mx/=sqrt(mx*mx+my*my)/5;
	////		    my/=sqrt(mx*mx+my*my)/5;
	////
	////		    mx+=bx;
	////		    my+=by;
	////
	////		    Vector2D m(mx,my);
	////                    agent->setIntention( new IntentionGetBall(m,	// the point 2-far from the ball in our goal direction
	////                                         ServerParam::i().maxDashPower(),
	////                                         0.9,
	////                                         5,
	////                                         wm.time() ) );
	////                    return true;
	////                }
	////
	////            }
	////        }
	////
	////    }
	//
	//
	//	//get active predicates the agent is preserving
	//		//do we have to work on a new predicate?
	//		//PredicateComplex::predicate toAdd=PredicateComplex::none;
	//
	//self_min = wm.interceptTable()->selfReachCycle();
	//mate_min = wm.interceptTable()->teammateReachCycle();
	//int opp_min = wm.interceptTable()->opponentReachCycle();
	int our_min = std::min( self_min, mate_min );

	dlog.addText( Logger::TEAM,__FILE__": PC -- possession me %d mate %d opp %d", self_min,mate_min,opp_min );


	//we have possession
	if (wm.self().unum()>4 &&
			((our_min <= opp_min - 2  || /*M_domain_graph.currentTstate()!=0 &&*/ M_domain_graph.currentTstate()!=13)
			||
			(our_min <= opp_min && wm.ball().pos().x>14))
	)
	{
		dlog.addText( Logger::TEAM,__FILE__": PC -- with possession" );
		//			dlog.addText( Logger::TEAM,__FILE__": PC -- predicates: %d %d %d",
		//					M_predicate_complex.o_activePredicates[0],
		//					M_predicate_complex.o_activePredicates[1],
		//					M_predicate_complex.o_activePredicates[2]);

		M_predicate_complex.o_activePredicates.clear();
		M_predicate_complex.o_activePredicates.push_back(PredicateComplex::open);
		M_predicate_complex.o_activePredicates.push_back(PredicateComplex::contact);
		M_predicate_complex.o_activePredicates.push_back(PredicateComplex::position);

		//			int s=M_predicate_complex.o_activePredicates.size();
		//			if (s<3)
		//			{
		//				//we can consider another predicate
		//				//choose at random now
		//				if (!((s>0 && M_predicate_complex.o_activePredicates[0]==PredicateComplex::open) ||
		//						(s>1 && M_predicate_complex.o_activePredicates[1]==PredicateComplex::open)))
		//					//add open
		//					toAdd= PredicateComplex::open;
		//				else if (!((s>0 && M_predicate_complex.o_activePredicates[0]==PredicateComplex::position) ||
		//						(s>1 && M_predicate_complex.o_activePredicates[1]==PredicateComplex::position)))
		//					//add position
		//					toAdd= PredicateComplex::position;
		//				else if (!((s>0 && M_predicate_complex.o_activePredicates[0]==PredicateComplex::contact) ||
		//						(s>1  && M_predicate_complex.o_activePredicates[1]==PredicateComplex::contact)))
		//					//add contact
		//					toAdd= PredicateComplex::contact;
		//				else if (!((s>0 && M_predicate_complex.o_activePredicates[0]==PredicateComplex::ball) ||
		//						(s>1  && M_predicate_complex.o_activePredicates[1]==PredicateComplex::ball)))
		//					//add ball
		//					toAdd= PredicateComplex::ball;
		//
		//			}
		//and find a pseudo-optimal point of the predicates
		std::list<Vector2D> targets;
		M_predicate_complex.generateTargets(wm,targets,true/*, toAdd*/);

		if (targets.size()>0)
		{
			//						dlog.addText( Logger::TEAM,__FILE__": PC -- targets available for %d, %d, %d ",
			//								M_predicate_complex.o_activePredicates.size()>0?M_predicate_complex.o_activePredicates[0]:-1,
			//										M_predicate_complex.o_activePredicates.size()>1?M_predicate_complex.o_activePredicates[1]:-1,
			//												M_predicate_complex.o_activePredicates.size()>2?M_predicate_complex.o_activePredicates[2]:-1);
			//
			//						int self_min = wm.interceptTable()->selfReachCycle();
			//						int mate_min = wm.interceptTable()->teammateReachCycle();
			//						int opp_min = wm.interceptTable()->opponentReachCycle();
			//						int our_min = std::min( self_min, mate_min );

			//complex 1 -- hold, dribble, pass
			//						if (wm.self().isKickable() || ( our_min <= opp_min - 2  &&  self_min <= mate_min - 2))
			//						{
			//
			//							//should clear?
			//							if(wm.ball().pos().dist(ServerParam::i().ourTeamGoalPos())<20 && wm.getDistOpponentNearestToBall(100,true)<5)
			//								if(Body_ClearBall2009().execute(agent))
			//									return true;
			//
			//							//if you have to pass, pass
			//							//have to pass when only contact is available or when other vertices start to disappear
			//							//if (M_predicate_complex.o_activePredicates[0]==PredicateComplex::contact)
			//								/*(M_predicate_complex.activePredicates.size()>0 && M_predicate_complex.activePredicates[0]==PredicateComplex::contact )
			//			                    ||
			//			                    (M_predicate_complex.activePredicates.size()>1 && M_predicate_complex.activePredicates[1]==PredicateComplex::contact)
			//			                    ||
			//			                    (M_predicate_complex.activePredicates.size()>2 && M_predicate_complex.activePredicates[2]==PredicateComplex::contact)*/
			//								// )
			//							{
			//								//    					Vector2D & pass_target = targets.front();
			//								dlog.addText( Logger::TEAM,
			//										__FILE__" (PredicateComplex Bhv_ChainAction) pass"); //(%.1f %.1f)
			//								int my_num=wm.self().unum();
			//
			//								if (wm.allTeammates().size()>1 && M_predicate_complex.pass_mate!=-1 &&
			//										(M_predicate_complex.pass_mate==0 || wm.teammate(M_predicate_complex.pass_mate)))// wm.getTeammateNearestToSelf(100,true))
			//								{
			//									if(M_predicate_complex.pass_mate!=0)
			//									{
			//										int mate_num=M_predicate_complex.pass_mate;//(wm.getTeammateNearestToSelf(100,true))->unum();
			//										Vector2D rec_pos(0,0);//=wm.getTeammateNearestToSelf(1)->pos();
			//										rcsc::Pass pass(my_num,mate_num,wm.teammate(M_predicate_complex.pass_mate)->pos()
			//												,0.2*wm.teammate(M_predicate_complex.pass_mate)->pos().dist(wm.self().pos())
			//												,10,1,false,"");
			//
			//										Bhv_PassKickFindReceiver passAction(M_domain_graph, M_predicate_complex);//,(CooperativeAction&)pass);
			//										if (passAction.execute(agent))/*,(CooperativeAction&)pass))*///(CooperativeAction::ConstPtr)(&pass)).execute(agent))
			//										{
			//											//std::cout<<"pass to mate"<<M_predicate_complex.pass_mate<<std::endl;//<<wm.getTeammateNearestToSelf(1)->unum()<<std::endl;
			//											return true;
			//										}
			//										else
			//										{
			//											std::cout<<" pass failed"<<std::endl;
			//										}
			//									}
			//									else //contact with goal
			//									{
			//										dlog.addText( Logger::TEAM,__FILE__" (PredicateComplex Bhv_BasicMove) shoot");
			//
			//										if(Bhv_Shoot2008().execute(agent))
			//											return true;
			//									}
			//
			//								}
			//							}
			//							//else	hold, dribble
			//							{
			//								Vector2D & dribble_target = targets.front();
			//								if (rand()%100 >50) dribble_target=targets.back();
			//
			//								dlog.addText( Logger::TEAM,
			//										__FILE__" (PredicateComplex Bhv_ChainAction) dribble target=(%.1f %.1f)",
			//										dribble_target.x, dribble_target.y );
			//
			//
			//								int max_speed=ServerParam::i().DEFAULT_PLAYER_SPEED_MAX+PlayerParam::i().playerSpeedMaxDeltaMax();
			//								double dist_target=wm.self().pos().dist(dribble_target);
			//
			//								//reaching the target in three steps?
			//								double vel=dist_target/(3+(dist_target/max_speed));
			//
			//
			//								agent->setIntention(
			//										new rcsc::IntentionDribble2008(dribble_target,0.9,0,3,ServerParam::i().maxDashPower()*0.7,false,wm.time())
			//								);
			//
			//								Vector2D me_targ=dribble_target-wm.self().pos();
			//								double turn_ang=fabs(me_targ.dir().degree()-wm.self().body().degree());
			//								if(turn_ang>180) turn_ang=360-turn_ang;
			//
			//
			//								if ( Bhv_NormalDribble(Dribble(wm.self().unum(),dribble_target,vel,1,turn_ang>15?1:0,1/*+(dist_target/max_speed)*/)).execute( agent ) )
			//								{
			//									return true;
			//								}
			//
			//
			//
			//								//					agent->setIntention(new IntentionDribble2008(dribble_target,10,1,3,ServerParam::i().MAX_DASH_POWER,false,wm.time()));
			//								//
			//								//					if ( Bhv_NormalDribble(new Dribble(wm.self().unum(),dribble_target,
			//								//							0.5*ServerParam::DEFAULT_PLAYER_SPEED_MAX,//0.15*wm.self().pos().dist(dribble_target),
			//								//							1,1,10).execute( agent ) )
			//								//					{
			//								//						return true;
			//								//					}
			//							}
			//
			//						}

			//						else //complex 2 -- support, receive, getBall
			{
				//				if(M_domain_graph.currentTstate()!=wm.self().unum())
				{
					Vector2D & target_point = targets.front();

					Vector2D head_goal=ServerParam::i().theirTeamGoalPos()-wm.self().pos();
					double safe_x=wm.offsideLineX()-2;
					head_goal.setLength(safe_x-target_point.x);


					if(wm.self().unum()==10)
						target_point=target_point+head_goal;

					dlog.addText( Logger::TEAM,
							__FILE__" (PredicateComplex Bhv_ChainAction) move target=(%.1f %.1f)",
							target_point.x, target_point.y );

					const double dash_power =ServerParam::i().MAX_DASH_POWER;

					double dist_thr = wm.ball().distFromSelf() * 0.1;
					if ( dist_thr < 1.0 ) dist_thr = 1.0;

					dlog.addText( Logger::TEAM,
							__FILE__": Bhv_BasicMove target=(%.1f %.1f) dist_thr=%.2f",
							target_point.x, target_point.y,
							dist_thr );

					//        agent->debugClient().addMessage( "BasicMove%.0f", dash_power );
					//        agent->debugClient().setTarget( target_point );
					//        agent->debugClient().addCircle( target_point, dist_thr );

					if ( ! Body_GoToPoint( target_point, dist_thr, dash_power
					).execute( agent ) )
					{
						Body_TurnToBall().execute( agent );
					}

					if ( wm.existKickableOpponent()
							&& wm.ball().distFromSelf() < 18.0 )
					{
						agent->setNeckAction( new Neck_TurnToBall() );
					}
					else
					{
						agent->setNeckAction( new Neck_TurnToBallOrScan() );
					}

					return true;

					//agent->setIntention(new IntentionGoToPoint(move_target,ServerParam::i().maxDashPower()*0.7,0.9,3,wm.time()));

					//Body_GoToPoint2010 gt( move_target,0.9,ServerParam::i().MAX_DASH_POWER);
					//						agent->setIntention(new IntentionGoToPoint(move_target,ServerParam::i().maxDashPower()*0.7,0.9,3,wm.time()));
					//
					//						if(Body_GoToPoint(move_target,0.9,ServerParam::i().maxDashPower()*0.7).execute(agent))
					//							return true;
				}

				//										else	//I had the ball before, the opponent is close -- fight for the ball
				//										{
				//											const Vector2D & move_target = wm.ball().inertiaPoint(3);
				//											dlog.addText( Logger::TEAM,
				//													__FILE__" (PredicateComplex Bhv_ChainAction) getBall target=(%.1f %.1f)",
				//													move_target.x, move_target.y );
				//
				//											agent->setIntention(new IntentionGoToPoint(move_target,ServerParam::i().maxDashPower(),0.9,3,wm.time()));
				//
				//											//Body_GoToPoint2010 gt( move_target,0.9,ServerParam::i().MAX_DASH_POWER);
				//											//agent->setIntention(new IntentionGoToPoint(move_target,ServerParam::i().maxDashPower()*0.7,0.9,3,wm.time()));
				//
				//											if(wm.ball().distFromSelf()<2)
				//											{
				//												if(Body_TackleToPoint(wm.ball().pos(),0.3,50).execute(agent))
				//													return true;
				//											}
				//											else
				//											{
				//												if(Body_GoToPoint(move_target,0.9,ServerParam::i().maxDashPower()).execute(agent))
				//													return true;
				//											}
				//										}



			}
		}
		//								else dlog.addText( Logger::TEAM,__FILE__": PC -- targets NOT available for %d, %d, %d ",
		//										(M_predicate_complex.o_activePredicates.size()>0?M_predicate_complex.o_activePredicates[0]:-1),
		//										(M_predicate_complex.o_activePredicates.size()>1?M_predicate_complex.o_activePredicates[1]:-1),
		//										(M_predicate_complex.o_activePredicates.size()>2?M_predicate_complex.o_activePredicates[2]:-1));

	}

	else	//we do not have possession -- defence
	{

		if(wm.self().unum()>4) dlog.addText( Logger::TEAM,__FILE__": PC -- no possession" );

		M_predicate_complex.d_activePredicates.clear();
		M_predicate_complex.d_activePredicates.push_back(PredicateComplex::marked);
		M_predicate_complex.d_activePredicates.push_back(PredicateComplex::blocked);
		//M_predicate_complex.d_activePredicates.push_back(PredicateComplex::supported);
//		int s=M_predicate_complex.d_activePredicates.size();
//		if (s<3)
//		{
//			//we can consider another predicate
//			//choose at random now
//			if (!((s>0 && M_predicate_complex.d_activePredicates[0]==PredicateComplex::blocked ) ||
//					(s>1 && M_predicate_complex.d_activePredicates[1]==PredicateComplex::blocked)))
//				//add blocked
//				toAdd= PredicateComplex::blocked;
//			else if (!((s>0 && M_predicate_complex.d_activePredicates[0]==PredicateComplex::marked ) ||
//					(s>1  && M_predicate_complex.d_activePredicates[1]==PredicateComplex::marked)))
//				//add marked
//				toAdd= PredicateComplex::marked;
//			else if (!((s>0 && M_predicate_complex.d_activePredicates[0]==PredicateComplex::supported ) ||
//					(s>1  && M_predicate_complex.d_activePredicates[1]==PredicateComplex::supported)))
//				//add supported
//				toAdd= PredicateComplex::supported;
//			else if (!((s>0 && M_predicate_complex.d_activePredicates[0]==PredicateComplex::tacklable ) ||
//					(s>1  && M_predicate_complex.d_activePredicates[1]==PredicateComplex::tacklable)))
//				//add tacklable
//				toAdd= PredicateComplex::tacklable;
//
//		}

//		//    			int self_min = wm.interceptTable()->selfReachCycle();
//		//    			int mate_min = wm.interceptTable()->teammateReachCycle();
//		//    			int opp_min = wm.interceptTable()->opponentReachCycle();
//		//    			int our_min = std::min( self_min, mate_min );
//		//
//		//    			if (wm.self().isKickable() || ( our_min <= opp_min - 2  &&  self_min <= mate_min - 2))
//		//    			{   //complex 1 -- hold, dribble, pass
//
//		//		//if you have to tackle, tackle
//		//		if (M_predicate_complex.d_activePredicates[0]==PredicateComplex::tacklable  ||
//		//				M_predicate_complex.d_activePredicates[1]==PredicateComplex::tacklable ||
//		//				M_predicate_complex.d_activePredicates[2]==PredicateComplex::tacklable)
//		//		{
//
//		//    					int my_num=wm.self().unum();
//
//		//	if (wm.self().distFromBall()<5 || (wm.getOpponentNearestToBall(100,true) && ((wm.getOpponentNearestToBall(100,true))->pos()).dist(wm.self().pos())<5 ))
////		{
////			dlog.addText( Logger::TEAM,__FILE__" (PredicateComplex Bhv_ChainAction) tackle");
////			Vector2D opp_pos=wm.ball().pos();//(wm.getOpponentNearestToBall(100,true))->pos();
////			dlog.addText( Logger::TEAM,
////					__FILE__"opponent position is (%f,%f), distBall %.1f, distOpp %.1f",opp_pos.x,opp_pos.y,wm.self().distFromBall(),wm.getDistOpponentNearestToBall(100,true));
////			//    						int mate_num=M_predicate_complex.pass_mate;//(wm.getTeammateNearestToSelf(100,true))->unum();
////			//    						Vector2D rec_pos(0,0);//=wm.getTeammateNearestToSelf(1)->pos();
////			Body_TackleToPoint tckl(opp_pos, 0.2,50);
////
////			if(tckl.execute(agent))
////			{
////				dlog.addText( Logger::TEAM,__FILE__"tackling ball holder");
////				return true;
////			}
////			//pass(my_num,mate_num,wm.teammate(M_predicate_complex.pass_mate)->pos(),10,10,1,false,"");
////
////			//    						Bhv_PassKickFindReceiver passAction(M_domain_graph);//,(CooperativeAction&)pass);
////			//    						if (passAction.execute(agent,(CooperativeAction&)pass))//(CooperativeAction::ConstPtr)(&pass)).execute(agent))
////			//    						{
////			//    							std::cout<<"pass to mate"<<M_predicate_complex.pass_mate<<std::endl;//<<wm.getTeammateNearestToSelf(1)->unum()<<std::endl;
////			//    							return true;
////			//    						}
////			else
////			{
////				dlog.addText( Logger::TEAM,__FILE__" tackle failed");
////			}
////		}
//		//}
//
////		dlog.addText( Logger::TEAM,__FILE__" no tackle");
//		//else	block, mark, support
//
//
//
////		if(self_min<=mate_min && self_min<=opp_min)
////		{
////			if(Body_Intercept2009().execute(agent))
////			{
////				dlog.addText( Logger::TEAM,__FILE__" Intercepting ball");
////				return true;
////			}
////		}


		//and find a pseudo-optimal point of the predicates
		std::list<Vector2D> targets;
		M_predicate_complex.generateTargets(wm,targets,false);//, toAdd);
		Vector2D move_target(-40,0);
		if(wm.getOpponentNearestToSelf(100,true))move_target=wm.getOpponentNearestToSelf(100,true)->pos();

//		if (targets.size()>0)
//		{
////			dlog.addText( Logger::TEAM,__FILE__": PC -- targets available  for %d, %d, %d ",
////					M_predicate_complex.d_activePredicates.size()>0?M_predicate_complex.d_activePredicates[0]:-1,
////							M_predicate_complex.d_activePredicates.size()>1?M_predicate_complex.d_activePredicates[1]:-1,
////									M_predicate_complex.d_activePredicates.size()>2?M_predicate_complex.d_activePredicates[2]:-1);
//
//
//
//			move_target= targets.front();
//			if (rand()%100 >50) move_target=targets.back();
//		}
//		else
//		{
//			dlog.addText( Logger::TEAM,__FILE__" targets NOT available");
//		}
//
//
		int marked_opp=-1;
		M_predicate_complex.instance().blocking=false;
//		//s=(int)M_predicate_complex.d_activePredicates.size();
////
////		if((s>0 && M_predicate_complex.d_activePredicates[0]==PredicateComplex::marked)
////				|| (s>1 && M_predicate_complex.d_activePredicates[1]==PredicateComplex::marked)
////				|| (s>2 && M_predicate_complex.d_activePredicates[2]==PredicateComplex::marked))
		{

			//find an unmarked opponent that may be dangerous
			Vector2D goal_c=ServerParam::instance().ourTeamGoalPos();
			Vector2D me_ball=wm.ball().pos()-wm.self().pos();
			Vector2D me_goal=ServerParam::i().ourTeamGoalPos()-wm.self().pos();

			Vector2D opp_pos=goal_c;	//if no specific opponent, mark the goal

			double dist;
			//const PlayerObject * opp=wm.getOpponentNearestToSelf(100,false);

			//opp_pos=opp->pos(); 	//take that as the marked
			//marked_opp=-1;//opp->unum();

			const PlayerObject * mate;

			AbstractPlayerCont opps=wm.allOpponents();
			std::sort( opps.begin(),opps.end(),PtrGoalBallDistCmp() );


			//			for ( PlayerPtrCont::const_iterator o = (wm.opponentsFromSelf()).begin();
			//			                  o != (wm.opponentsFromSelf()).end();
			//			                  ++o )
			for ( AbstractPlayerCont::const_iterator o = opps.begin();o != opps.end();++o )
			{
				dlog.addText( Logger::TEAM,__FILE__" OppCont -- %d in %.2f %.2f",(*o)->unum(),(*o)->pos().x,(*o)->pos().y);
				if((*o)->unum()!=-1 && (*o)->pos().x<0)	//attacking player
				{
					if(!M_predicate_complex.Marked((*o)->unum()))	//not marked already
					{
						mate=wm.getTeammateNearestTo((const PlayerObject*)(*o),100,&dist);
						int op=0;
						if(mate)
						{
							for(;op<=11;op++)
							{
								if(op==(*o)->unum())continue;
								if(M_predicate_complex.Marker(op)==mate->unum())		//mate is busy with op
									break;
							}
							if(op!=12 /*mate is engaged*/ || dist>=wm.self().pos().dist((*o)->pos()))	//free but I'm closer -- mark o
							{
								opp_pos=(*o)->pos(); 	//take that as the marked
								marked_opp=(*o)->unum();

								break;
							}
						}
					}
					else if((*o)->unum()!=-1)
					{
						dlog.addText( Logger::TEAM,__FILE__" marking -- %d already marked by %d",(*o)->unum(),M_predicate_complex.Marker((*o)->unum()));

					}
				}
				else
					dlog.addText( Logger::TEAM,__FILE__" marking -- %d too far to be marked");


			}


			dlog.addText( Logger::TEAM,__FILE__" marking %d in (%.1f,%.1f)",marked_opp,opp_pos.x,opp_pos.y);

			//dlog.addText( Logger::TEAM,__FILE__" nearest opp %d, nearest mate to opp %d",opp->unum(),mate->unum());


			//			if(dist>=wm.getDistOpponentNearestToSelf(100,false))	//I'm closer
			//			{
			//				//	)->unum()==wm.self().unum())
			//				//I'm the closest to that opponent, I should be marking him
			//				//getOpponentNearestToBall(100,true))
			//				opp_pos=opp->pos(); 	//take that as the marked
			//				marked_opp=opp->unum();
			//				dlog.addText( Logger::TEAM,__FILE__" marking %d in (%.1f,%.1f)",opp->unum(),opp->pos().x,opp->pos().y);
			//
			//			}
			//			else
			//			{
			//				const PlayerObject * opp2=wm.getOpponentNearestTo(opp,100,&dist);
			//				if(opp2)
			//				{
			//					opp_pos=opp2->pos();
			//					dlog.addText( Logger::TEAM,__FILE__" second nearest opp %d",opp2->unum());
			//
			//					marked_opp=opp2->unum();
			//					dlog.addText( Logger::TEAM,__FILE__" marking %d in (%.1f,%.1f)",opp2->unum(),opp2->pos().x,opp2->pos().y);
			//				}
			//			}

			Vector2D me_opp=opp_pos-wm.self().pos();
			Vector2D opp_ball=wm.ball().pos()-opp_pos;
			Vector2D opp_goal=ServerParam::i().ourTeamGoalPos()-opp_pos;


			double angle=0;
			double diff=0;

			bool opp_kickable=false;
			if(marked_opp>=1 && marked_opp<=11)
				opp_kickable=(/*opp_goal.norm()<18  &&*/ ((PlayerObject*)(wm.opponent(marked_opp)))->isKickable()  || (wm.getOpponentNearestToBall(100,true) && wm.getOpponentNearestToBall(100,true)->unum()==marked_opp));

			if(!opp_kickable )
				diff=fabs(me_ball.dir().degree()-me_opp.dir().degree());
			else
				diff=fabs(me_goal.dir().degree()-me_opp.dir().degree());


			if(diff>180)
			{
				diff=360-diff;
			}

			dlog.addText( Logger::TEAM,__FILE__" marking -- me_b %.1f me_o %.1f diff/? %.1f - kickable? %d",me_ball.dir().degree(),me_opp.dir().degree(),diff/(opp_kickable?2:3),opp_kickable?1:0);

			if (
					(!opp_kickable &&
							(
									(me_ball.dir().degree()>me_opp.dir().degree() && !(me_ball.dir().degree()>0 && me_opp.dir().degree()<0 && fabs(me_ball.dir().degree()-me_opp.dir().degree())>180))
									||
									(me_ball.dir().degree()<0 && me_opp.dir().degree()>0 && fabs(me_ball.dir().degree()-me_opp.dir().degree())>180)
							)
					)
					||
					(opp_kickable &&
							(
									(me_goal.dir().degree()>me_opp.dir().degree() && !(me_goal.dir().degree()>0 && me_opp.dir().degree()<0 && fabs(me_goal.dir().degree()-me_opp.dir().degree())>180))
									||
									(me_goal.dir().degree()<0 && me_opp.dir().degree()>0 && fabs(me_goal.dir().degree()-me_opp.dir().degree())>180)
							)
					)
			)
				//greater or b is smaller but still need to add
				angle=me_opp.dir().degree()+diff/(opp_kickable?2:3);
			else
				angle=me_opp.dir().degree()-diff/(opp_kickable?2:3);



			//			//the angles have different signs
			//			if (me_ball.dir().degree()>me_opp.dir().degree()) 	//positive
			//			{
			//				if(fabs(me_ball.dir().degree())<fabs(me_opp.dir().degree())) //smaller
			//					angle=me_ball.dir().degree()+diff/4;
			//				else
			//					angle=me_opp.dir().degree()-diff/4;
			//			}
			//			else
			//			{
			//				if(fabs(me_ball.dir().degree())<fabs(me_opp.dir().degree())) //smaller
			//					angle=me_ball.dir().degree()-diff/4;
			//				else
			//					angle=me_opp.dir().degree()+diff/4;
			//			}
			//		}
			//		else
			//		{
			//			if(me_ball.dir().degree()*me_opp.dir().degree() >0) 	//of same sign
			//			{
			//				if(me_ball.dir().degree()<me_opp.dir().degree()) //smaller
			//					angle=me_ball.dir().degree()+diff/4;
			//				else
			//					angle=me_opp.dir().degree()+diff/4;
			//			}
			//			else	//different signs
			//			{
			//				if(me_ball.dir().degree()<me_opp.dir().degree()) //negative
			//					angle=me_ball.dir().degree()+diff/4;
			//				else
			//					angle=me_opp.dir().degree()+diff/4;
			//			}

			//			}


			double r=0;
			if(!opp_kickable)
					r=me_ball.norm()*me_opp.norm()*(1-(pow(opp_ball.norm(),2)/std::pow((me_ball.norm()+me_opp.norm()),2)));
					//r=me_ball.norm()-opp_ball.norm()+(M_target_player->unum()!=-1?0.33:0.37)*opp_ball.norm();
				else
					r=me_goal.norm()*me_opp.norm()*(1-(pow(opp_goal.norm(),2)/std::pow((me_opp.norm()+me_goal.norm()),2)));
					//r=me_goal.norm()-opp_goal.norm()+0.33*opp_goal.norm();

			if(r<0)r=0;
//			{
//				angle+=180;
//				if(angle>180)angle-=360;
//				r*=-1;
//			}

			dlog.addText( Logger::TEAM,__FILE__" marking -- angle %.1f r %.1f",angle,r);

			move_target.setPolar(r,AngleDeg(angle));

			move_target=move_target+wm.self().pos();


			Vector2D cover=opp_goal;
			cover.setLength(2);

			if(marked_opp==-1 && wm.getOpponentNearestToSelf(100,true) && wm.ball().pos().x<0)
							move_target.y=0.5*(move_target.y+wm.getOpponentNearestToSelf(100,true)->pos().y);


			//if the opponent is not in shooting range, keep behind him
			if(opp_pos.dist(ServerParam::i().ourTeamGoalPos())>15)
				//dribble_target.x=dribble_target.x-2;
				move_target=move_target+cover;


			dlog.addText( Logger::TEAM,__FILE__" marking target=(%.1f %.1f) -- angle %.1f",
					move_target.x, move_target.y,angle );
		}
//
//
////		if((s>0 && M_predicate_complex.d_activePredicates[0]==PredicateComplex::blocked)
////				|| (s>1 && M_predicate_complex.d_activePredicates[1]==PredicateComplex::blocked)
////				|| (s>2 && M_predicate_complex.d_activePredicates[2]==PredicateComplex::blocked))
//		{

		if(M_predicate_complex.isBlocked(wm.self().pos(),wm)) //only if not blocked already
				dlog.addText( Logger::TEAM,__FILE__" readily blocked");

			else
			{
				//if somebody is going to block already
				if(!M_predicate_complex.Marked(0) ||
						M_predicate_complex.Marker(0)==-1 ||
						( wm.teammate(M_predicate_complex.Marker(0))
								&& wm.teammate(M_predicate_complex.Marker(0))->pos().dist(ServerParam::i().ourTeamGoalPos()) > wm.self().pos().dist(ServerParam::i().ourTeamGoalPos())))
				{
					Vector2D goal_c=ServerParam::instance().ourTeamGoalPos();
					Vector2D me_goal=goal_c-wm.self().pos();

					Vector2D opp_pos=wm.ball().pos();

					if(wm.getOpponentNearestToBall(100,true))
					{
						opp_pos=opp_pos+(wm.getOpponentNearestToBall(100,true))->pos();
						opp_pos.scale(0.5);	//the point between the player and the ball
					}
					double dist=0;
					const rcsc::PlayerObject * mate=wm.getTeammateNearestTo(wm.getOpponentNearestToBall(100,true),100,&dist);
					if(dist>=wm.self().pos().dist(opp_pos))	//I'm closer -- block
					{

						Vector2D me_opp=opp_pos-wm.self().pos();

						double angle=0;
						double diff=fabs(me_goal.dir().degree()-me_opp.dir().degree());


						dlog.addText( Logger::TEAM,__FILE__" blocking -- me_goal %.1f me_opp %.1f diff %.1f",me_goal.dir().degree(),me_opp.dir().degree(),diff);


						if(diff>180)
						{
							diff=360-diff;
							//the angles have different signs
							if (me_goal.dir().degree()>me_opp.dir().degree()) 	//positive
							{
								if(fabs(me_goal.dir().degree())<fabs(me_opp.dir().degree())) //smaller
									angle=me_goal.dir().degree()+diff/2;
								else
									angle=me_opp.dir().degree()-diff/2;
							}
							else
							{
								if(fabs(me_goal.dir().degree())<fabs(me_opp.dir().degree())) //smaller
									angle=me_goal.dir().degree()-diff/2;
								else
									angle=me_opp.dir().degree()+diff/2;
							}
						}
						else
						{
							if(me_goal.dir().degree()*me_opp.dir().degree() >0) 	//of same sign
							{
								if(me_goal.dir().degree()<me_opp.dir().degree()) //smaller
									angle=me_goal.dir().degree()+diff/2;
								else
									angle=me_opp.dir().degree()+diff/2;
							}
							else	//different signs
							{
								if(me_goal.dir().degree()<me_opp.dir().degree()) //negative
									angle=me_goal.dir().degree()+diff/2;
								else
									angle=me_opp.dir().degree()+diff/2;
							}

						}

						dlog.addText( Logger::TEAM,__FILE__" blocking target -- angle %.1f",angle);
						Vector2D opp_goal=ServerParam::i().ourTeamGoalPos()-opp_pos;

						double r=0;
						//r=me_goal.norm()-opp_goal.norm()+0.33*opp_goal.norm();
						r=me_goal.norm()*me_opp.norm()*(1-(pow(opp_goal.norm(),2)/std::pow((me_opp.norm()+me_goal.norm()),2)));

						if(r<0)r=0;

						dlog.addText( Logger::TEAM,__FILE__" marking -- angle %.1f r %.1f",angle,r);


						move_target.setPolar(r,AngleDeg(angle));
						move_target=move_target+wm.self().pos();
						dlog.addText( Logger::TEAM,__FILE__" blocking target=(%.1f %.1f)",
								move_target.x, move_target.y );
						M_predicate_complex.instance().blocking=true;
						marked_opp=-1;
					}
				}
				else
					dlog.addText( Logger::TEAM,__FILE__" blocking under way...");

			}
//		}


		if(wm.ball().pos().x>0)	//the ball is in their half
		{
			//tell defenders not to advance much
			if(wm.self().unum()<=4)
			{
				if(move_target.x>-15)
					move_target.x=std::min(-15,(int)move_target.x-7);
			}
		}


		dlog.addText( Logger::TEAM,
				__FILE__" (PredicateComplex Bhv_ChainAction) final moving target=(%.1f %.1f)",
				move_target.x, move_target.y );


		//dribble needs ball to be kickable

		//
		//    				Dribble d(wm.self().unum(),move_target,0.15*wm.self().pos().dist(move_target),1,1,10);
		//    				if ( Bhv_NormalDribble(d).execute( agent ) )
		//    				{
		//    					return true;
		//    				}
		//

		Vector2D me_targ=move_target-wm.self().pos();


		if(move_target.x>ServerParam::i().DEFAULT_PITCH_LENGTH/2)
			move_target.x=ServerParam::i().DEFAULT_PITCH_LENGTH/2-0.5;
		else if(move_target.x<-1*ServerParam::i().DEFAULT_PITCH_LENGTH/2)
			move_target.x=-1*ServerParam::i().DEFAULT_PITCH_LENGTH/2+0.5;

		if(move_target.y>ServerParam::i().DEFAULT_PITCH_WIDTH/2)
			move_target.y=ServerParam::i().DEFAULT_PITCH_WIDTH/2-0.5;
		else if(move_target.y<-1*ServerParam::i().DEFAULT_PITCH_WIDTH/2)
			move_target.y=-1*ServerParam::i().DEFAULT_PITCH_WIDTH/2+0.5;



		 double dist_thr = wm.ball().distFromSelf() * 0.1;
		    if ( dist_thr < 1.0 ) dist_thr = 1.0;




		//agent->doPointto(move_target.x, move_target.y);


		if(M_predicate_complex.instance().blocking)// 	marked_opp==-1)
			agent->setIntention(new IntentionBlock(3,wm.time()));
		else
		{
			if(marked_opp>=1 && marked_opp<=11)
			{
				agent->setIntention(new IntentionMark((const rcsc::PlayerObject *)(wm.opponent(marked_opp)),ServerParam::i().maxDashPower(),0.9,4,wm.time()));
				if(wm.opponent(marked_opp)->distFromSelf()<10)
					M_predicate_complex.marked_unum=marked_opp;
				else
					M_predicate_complex.marked_unum=-1;
			}
		}


//		if (! wm.existKickableTeammate()
//		                && ( self_min <= 3))
//		    {
//		            dlog.addText( Logger::TEAM,
//		                          __FILE__": intercept" );
//		            Body_Intercept().execute( agent );
//		            agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
//
//		            return true;
//		    }

		if(Body_GoToPoint(move_target,dist_thr, ServerParam::i().MAX_DASH_POWER).execute(agent))
			//if(agent->doDash(75,me_targ.dir()-wm.self().body()))
		{
			dlog.addText( Logger::TEAM,__FILE__" dashing to dribble target=(%.1f %.1f) using angle %f -   %f ",
					move_target.x, move_target.y,me_targ.dir().degree(),wm.self().body().degree());
			return true;
		}
//		else
//			{
//			dlog.addText( Logger::TEAM,__FILE__" goToPoint failed");
//	        Body_TurnToBall().execute( agent );
//
//			}

	    if ( wm.existKickableOpponent() && wm.ball().distFromSelf() < 18.0 )
	    {
	        agent->setNeckAction( new Neck_TurnToBall() );
	    }
	    else
	    {
	        agent->setNeckAction( new Neck_TurnToBallOrScan() );
	    }




	}

	return execute(agent);


}




//   /*
//     //-----------------------------------------------
//     // tackle
//     if ( Bhv_BasicTackle( 0.8, 80.0 ).execute( agent ) )
//     {
// 	return true;
//     }
//
//     const WorldModel & wm = agent->world();
//     /*--------------------------------------------------------*/
//     // chase ball
//     const int self_min = wm.interceptTable()->selfReachCycle();
//     const int mate_min = wm.interceptTable()->teammateReachCycle();
//     const int opp_min = wm.interceptTable()->opponentReachCycle();
//
//
//     if (! wm.existKickableTeammate()
// 	&& ( self_min <= 3
// 	|| ( self_min <= mate_min
// 	&& self_min < opp_min + 3 )
// 	)
//     )
//     {
// 	dlog.addText( Logger::TEAM,
// 		      __FILE__": intercept" );
// 	Body_Intercept().execute( agent );
// 	agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
//
// 	return true;
//     }
//
//     const Vector2D target_point = Strategy::i().getPosition( wm.self().unum() );
//     const double dash_power = Strategy::get_normal_dash_power( wm );
//
//     double dist_thr = wm.ball().distFromSelf() * 0.1;
//     if ( dist_thr < 1.0 ) dist_thr = 1.0;
//
// 		     dlog.addText( Logger::TEAM,
// 				   __FILE__": Bhv_BasicMove target=(%.1f %.1f) dist_thr=%.2f",
// 				   target_point.x, target_point.y,
// 				   dist_thr );
//
// 		     agent->debugClient().addMessage( "BasicMove%.0f", dash_power );
//     agent->debugClient().setTarget( target_point );
//     agent->debugClient().addCircle( target_point, dist_thr );
//
//     if ( ! Body_GoToPoint( target_point, dist_thr, dash_power
//     ).execute( agent ) )
//     {
// 	Body_TurnToBall().execute( agent );
//     }
//
//     if ( wm.existKickableOpponent()
// 	&& wm.ball().distFromSelf() < 18.0 )
//     {
// 	agent->setNeckAction( new Neck_TurnToBall() );
//     }
//     else
//     {
// 	agent->setNeckAction( new Neck_TurnToBallOrScan() );
//     }
//
//     return true;*/
//}
// /*
// //-----------------------------------------------
// // tackle
// if ( Bhv_BasicTackle( 0.8, 80.0 ).execute( agent ) )
// {
//     return true;
// }
//
// const WorldModel & wm = agent->world();
// /*--------------------------------------------------------*/
// // chase ball
// const int self_min = wm.interceptTable()->selfReachCycle();
// const int mate_min = wm.interceptTable()->teammateReachCycle();
// const int opp_min = wm.interceptTable()->opponentReachCycle();
//
// if ( ! wm.existKickableTeammate()
//     && ( self_min <= 3
//     || ( self_min <= mate_min
//     && self_min < opp_min + 3 )
//     )
// )
// {
//     dlog.addText( Logger::TEAM,
// 		  __FILE__": intercept" );
//     Body_Intercept().execute( agent );
//     agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
//
//     return true;
// }
//
// if ( wm.existKickableOpponent()
//     && wm.ball().distFromSelf() < 18.0 )
// {
//     agent->setNeckAction( new Neck_TurnToBall() );
// }
// else
// {
//     agent->setNeckAction( new Neck_TurnToBallOrScan() );
// }
//
// return true;
// }*/
